# (C) 2025 GoodData Corporation
"""This module defines type aliases intended to improve readability."""

from typing import TypeAlias

WorkspaceId: TypeAlias = str
DatasetId: TypeAlias = str

__all__ = ["WorkspaceId", "DatasetId"]
